from cv2 import cv2
import pickle
import numpy as np
import os


def save_calib_data(calib_data: object, calib_data_path: str):
    """Saves calibration data."""
    with open(calib_data_path, "wb") as f:
        pickle.dump(calib_data, f)


def load_calib_data(calib_data_path: str):
    """Loads calibration data."""
    with open(calib_data_path, "rb") as f:
        calib_data = pickle.load(f)
    return calib_data


def draw_axis(img, corners, imgpts, thickness=5):
    """Draws xyz axis."""
    # NOTE: opencv in BGR order, B-z, G-y, R-x
    corner = tuple(corners[0].astype(int).ravel())
    img = cv2.line(
        img,
        corner,
        tuple(imgpts[0].astype(int).ravel()),
        (0, 0, 255),
        thickness=thickness,
    )
    img = cv2.line(
        img,
        corner,
        tuple(imgpts[1].astype(int).ravel()),
        (0, 255, 0),
        thickness=thickness,
    )
    img = cv2.line(
        img,
        corner,
        tuple(imgpts[2].astype(int).ravel()),
        (255, 0, 0),
        thickness=thickness,
    )
    return img


def collect_checker_board_images(camera, save_dir):
    """Collects checkerboard images."""

    print("Type [c] to capture, [q] to exit.")
    count = 1
    while True:
        _, frame = camera.read()
        cv2.imshow("Frame", frame)

        key = cv2.waitKey(1)
        if key == ord("q"):
            break
        elif key == ord("c"):
            cv2.imwrite(os.path.join(save_dir, f"image_{count}.png"), frame)
            print(f"Saved data #{count}")
            count += 1

    camera.release()
    cv2.destroyAllWindows()


def calibrate_camera_checkerboard(images, cols, rows, square_size, verbose=True):
    """Calibrates camera to get camera matrix and distortion coefficients."""

    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)

    # prepare object points
    objp = np.zeros((cols * rows, 3), np.float32)
    objp[:, :2] = np.mgrid[0:cols, 0:rows].T.reshape(-1, 2)
    objp = objp * square_size

    # arrays to store object points and image points from all the images
    objpoints = []
    imgpoints = []

    for img in images:
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        ret, corners = cv2.findChessboardCorners(gray, (cols, rows), None)

        if ret:
            objpoints.append(objp)

            corners = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1), criteria)
            imgpoints.append(corners)

            out_img = img.copy()
            if verbose:
                cv2.drawChessboardCorners(out_img, (cols, rows), corners, ret)
                cv2.imshow("Calibration", out_img)
                cv2.waitKey(0)

    cv2.destroyAllWindows()

    rmse, camera_matrix, dist_coeffs, _, _ = cv2.calibrateCamera(
        objpoints, imgpoints, gray.shape[::-1], None, None
    )

    return rmse, camera_matrix, dist_coeffs


def find_aruco_pose(frame, camera_matrix, dist_coeffs, marker_length):
    """Finds aruco marker pose."""
    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    aruco_dict = cv2.aruco.Dictionary_get(cv2.aruco.DICT_7X7_50)
    parameters = cv2.aruco.DetectorParameters_create()
    corners, ids, _ = cv2.aruco.detectMarkers(
        gray,
        aruco_dict,
        parameters=parameters,
        cameraMatrix=camera_matrix,
        distCoeff=dist_coeffs,
    )
    R_target2cam, t_target2cam = None, None
    if np.all(ids is not None):
        for i in range(len(ids)):
            rvec, tvec, _ = cv2.aruco.estimatePoseSingleMarkers(
                corners=corners[i],
                markerLength=marker_length,
                cameraMatrix=camera_matrix,
                distCoeffs=dist_coeffs,
            )
            cv2.aruco.drawDetectedMarkers(frame, corners)
            cv2.aruco.drawAxis(
                image=frame,
                cameraMatrix=camera_matrix,
                distCoeffs=dist_coeffs,
                rvec=rvec,
                tvec=tvec,
                length=50,
            )
            R_target2cam, _ = cv2.Rodrigues(rvec)
            t_target2cam = tvec.squeeze()

    return R_target2cam, t_target2cam


def find_checkerboard_pose(frame, camera_matrix, dist_coeffs, cols, rows, square_size):
    """Finds checkerboard pose."""
    R_target2cam, t_target2cam = None, None

    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    objp = np.zeros((cols * rows, 3), np.float32)
    objp[:, :2] = np.mgrid[0:cols, 0:rows].T.reshape(-1, 2)
    objp = objp * square_size

    axis = np.float32(
        [[3 * square_size, 0, 0], [0, 3 * square_size, 0], [0, 0, -3 * square_size]]
    ).reshape(-1, 3)

    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

    ret, corners = cv2.findChessboardCorners(gray, (cols, rows), None)

    if ret:
        # refine corners
        corners = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1), criteria)

        # find pose
        ret, rvec, tvec = cv2.solvePnP(objp, corners, camera_matrix, dist_coeffs)
        R_target2cam, _ = cv2.Rodrigues(rvec)
        t_target2cam = tvec.squeeze()

        imgpts, _ = cv2.projectPoints(axis, rvec, tvec, camera_matrix, dist_coeffs)
        frame = draw_axis(frame, corners, imgpts)

    return R_target2cam, t_target2cam


def collect_eye_hand_data(
    camera,
    camera_matrix,
    dist_coeffs,
    marker_length,
    robot,
    aruco,
    cols,
    rows,
    square_size,
):
    """Collects eye hand calibration data."""
    target_poses = []
    robot_ee_poses = []
    count = 0
    while True:
        _, frame = camera.read()
        if aruco:
            R_target2cam, t_target2cam = find_aruco_pose(
                frame=frame,
                camera_matrix=camera_matrix,
                dist_coeffs=dist_coeffs,
                marker_length=marker_length,
            )
        else:
            R_target2cam, t_target2cam = find_checkerboard_pose(
                frame=frame,
                camera_matrix=camera_matrix,
                dist_coeffs=dist_coeffs,
                cols=cols,
                rows=rows,
                square_size=square_size,
            )
        cv2.imshow("frame", frame)

        key = cv2.waitKey(1)
        if key == ord("q"):
            break
        if key == ord("c"):
            if R_target2cam is None:
                print("None pose detected, try again")
            else:
                target_poses.append((R_target2cam, t_target2cam))

                robot.connect()
                xyzrpw = robot.get_curpos()
                robot_ee_poses.append(xyzrpw)
                count += 1
                print(f"Collected data: {count}")

    return target_poses, robot_ee_poses


def calibrate_eye_hand(
    R_gripper2base, t_gripper2base, R_target2cam, t_target2cam, eye_to_hand=True
):

    tts = []
    if eye_to_hand:
        # change coordinates from gripper2base to base2gripper
        R_base2gripper, t_base2gripper = [], []
        for R, t in zip(R_gripper2base, t_gripper2base):
            R_b2g = R.T
            t_b2g = -R_b2g @ t
            tts.append(t_b2g)
            R_base2gripper.append(R_b2g)
            t_base2gripper.append(t_b2g)

        # change parameters values
        R_gripper2base = R_base2gripper
        t_gripper2base = t_base2gripper

    # calibrate
    R, t = cv2.calibrateHandEye(
        R_gripper2base=R_gripper2base,
        t_gripper2base=t_gripper2base,
        R_target2cam=R_target2cam,
        t_target2cam=t_target2cam,
    )

    return R, t



from abc import ABC, abstractmethod
import traceback


class RobotApp(ABC):
    """Generic Robot App Interface.
    Use this interface to develop robotic apps.
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def configure(self):
        raise NotImplementedError("Should implement configure method!")

    @abstractmethod
    def _main(self, **kwargs):
        """Main method"""
        raise NotImplementedError("Should implement run method!")

    def run(self, **kwargs):
        """Execution method."""
        status = True
        message = "success"
        result = None
        try:
            result = self._main(**kwargs)
        except Exception as excp:
            status = False
            message = "".join(
                traceback.TracebackException.from_exception(excp).format()
            )
            result = None
        finally:
            return status, message, result



import numpy as np
from scipy.spatial.transform import Rotation


def Rt_to_H(R, t):
    """Converts 3x3 rotation matrix 3x1 translation vector into
    4x4 homogeneous transformation matrix."""
    H = np.eye(4)
    H[0:3, 0:3] = R
    H[0:3, 3] = t.squeeze()
    return H


def H_to_Rt(H):
    """Converts 4x4 homogeneous transformation matrix into
    3x3 rotation matrix and 3x1 translation vector."""
    return H[0:3, 0:3], H[0:3, 3]


def xyzrpw_to_H(xyzrpw):
    """Converts xyzrpw to 4x4 homogeneous transformation matrix."""
    R = Rotation.from_euler("xyz", xyzrpw[3:], degrees=True).as_matrix()
    H = Rt_to_H(R, xyzrpw[:3])
    return H


def H_to_xyzrpw(H):
    """Converts 4x4 homogeneous transformation matrix to xyzrpw."""
    R, t = H_to_Rt(H)
    rpw = Rotation.from_matrix(R).as_euler("xyz", degrees=True)
    return list(t.flatten()) + list(rpw.flatten())



from abc import ABC
import socket
import sys


class Robot(ABC):
    def __init__(
        self,
        robot_model: str,
        host: str,
        port: int = 18375,
        ee_DO_type: str = None,
        ee_DO_num: int = None,
        socket_timeout: int = 60,
    ):
        """[summary]

        Args:
            robot_model (str): Robot model: Fanuce, Kuka, etc.
            host (str): IP address of host.
            port (int): Port number. Defaults to 18735.
            ee_DO_type (str, optional): End-effector digital output type.
                                        Fanuc used RDO type. Defaults to None.
                                        Others may use DO type.
            ee_DO_num (int, optional): End-effector digital output number.
                                       Defaults to None.
            socket_timeout(int): Socket timeout in seconds. Defaults to 5 seconds.
        """
        super().__init__()

        self.robot_model = robot_model
        self.host = host
        self.port = port
        self.ee_DO_type = ee_DO_type
        self.ee_DO_num = ee_DO_num
        self.sock_buff_sz = 1024
        self.socket_timeout = socket_timeout
        self.comm_sock = None
        self.SUCCESS_CODE = 0
        self.ERROR_CODE = 1

    def __version__(self):
        print("MAPPDK Robot class v0.1.11")

    def handle_response(self, resp: str, verbose: bool = False):
        """Handles response from socket communication.

        Args:
            resp (str): Response string returned from socket.
            verbose (bool, optional): [description]. Defaults to False.

        Returns:
            tuple(int, str): Response code and response message.
        """
        code, msg = resp.split(":")
        code = int(code)

        if code == self.SUCCESS_CODE:
            if verbose:
                print(f"The instruction was successfully executed. Message: {msg}.")
        elif code == self.ERROR_CODE:
            raise Exception(msg)
        else:
            print(f"Something wrong: {msg}")
            sys.exit()

        return code, msg

    def connect(self) -> None:
        """Connects to the physical robot."""
        # create socket and connect
        self.comm_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.comm_sock.settimeout(self.socket_timeout)
        self.comm_sock.connect((self.host, self.port))
        resp = self.comm_sock.recv(self.sock_buff_sz).decode()
        _, _ = self.handle_response(resp, verbose=True)

    def disconnect(self) -> None:
        self.comm_sock.close()

    def send_cmd(self, cmd: str):
        """Sends command to a physical robot.

        Args:
            cmd (str): Command string.

        Returns:
            tuple(int, str): Response code and response message.
        """
        # end of command character
        cmd = cmd.strip() + "\n"

        # 1. send
        self.comm_sock.sendall(cmd.encode())

        # 2. wait for a result
        resp = self.comm_sock.recv(self.sock_buff_sz).decode()
        return self.handle_response(resp)

    def call_prog(self, prog_name: str) -> None:
        """Calls external program name in a physical robot.

        Args:
            prog_name ([str]): External program name.
        """
        cmd = f"mappdkcall:{prog_name}"
        self.send_cmd(cmd)

    def get_ins_power(self) -> float:
        """Gets instantaneous power consumption.

        Returns:
            float: Watts.
        """

        cmd = "ins_pwr"
        _, msg = self.send_cmd(cmd)

        # Fanuc returns in kW. Should be adjusted to other robots.
        ins_pwr = float(msg) * 1000

        return ins_pwr

    def get_curpos(self):
        """Gets current cartesian position of tool center point.

        Returns:
            list[float]: Current positions XYZWPR.
        """

        cmd = "curpos"
        _, msg = self.send_cmd(cmd)
        vals = [float(val.split("=")[1]) for val in msg.split(",")]
        return vals

    def get_curjpos(self):
        """Gets current joint values of tool center point.

        Returns:
            list[float]: Current joint values.
        """
        cmd = "curjpos"
        _, msg = self.send_cmd(cmd)
        vals = [float(val.split("=")[1]) for val in msg.split(",") if val != "j=none"]
        return vals

    def move(
        self,
        move_type: str,
        vals: list,
        velocity: int = 25,
        acceleration: int = 100,
        cnt_val: int = 0,
        linear: bool = False,
    ) -> None:
        """Moves robot.

        Args:
            move_type (str): Movement type (joint or pose).
            vals (list[real]): Position values.
            velocity (int, optional): Percentage or mm/s. Defaults to 25%.
            acceleration (int, optional): Percentage or mm/s^2. Defaults to 100%.
            cnt_val (int, optional): Continuous value for stopping. Defaults to 50.
            linear (boolean, optioal): Linear movement. Defaults to False.

        Raises:
            ValueError: raises if movement type is not one of ("movej", "movep")
        """

        # prepare velocity. percentage or mm/s
        # format: aaaa, e.g.: 0001%, 0020%, 3000 mm/s
        velocity = int(velocity)
        velocity = f"{velocity:04}"

        # prepare acceleration. percentage or mm/s^2
        # format: aaaa, e.g.: 0001%, 0020%, 0100 mm/s^2
        acceleration = int(acceleration)
        acceleration = f"{acceleration:04}"

        # prepare CNT value
        # format: aaa, e.g.: 001, 020, 100
        cnt_val = int(cnt_val)
        assert 0 <= cnt_val <= 100, "Incorrect CNT value."
        cnt_val = f"{cnt_val:03}"

        if move_type == "joint":
            cmd = "movej"
        elif move_type == "pose":
            cmd = "movep"
        else:
            raise ValueError("Incorrect movement type!")

        if linear:
            motion_type = 1
        else:
            motion_type = 0

        cmd += f":{velocity}:{acceleration}:{cnt_val}:{motion_type}:{len(vals)}"

        # prepare joint values
        for val in vals:
            vs = f"{abs(val):013.6f}"
            if val >= 0:
                vs = "+" + vs
            else:
                vs = "-" + vs
            cmd += f":{vs}"

        # call send_cmd
        self.send_cmd(cmd)

    def gripper(self, value: bool) -> None:
        """Opens/closes robot gripper.

        Args:
            value (boolean): True or False
        """
        if (self.ee_DO_type is not None) and (self.ee_DO_num is not None):
            value = "true" if value else "false"

            cmd = ""
            if self.ee_DO_type == "RDO":
                cmd = "setrdo"
            elif self.ee_DO_type == "DO":
                cmd = "setdo"
            else:
                raise ValueError("Wrong DO type!")

            cmd = cmd + f":{self.ee_DO_num }:{value}"
            self.send_cmd(cmd)
        else:
            raise ValueError("DO type or number is None!")

    def get_rdo(self, rdo_num: int):
        """Get RDO value.

        Args:
            rdo_num (int): RDO number.

        Returns:
            rdo_value: RDO value.
        """
        cmd = f"getrdo:{rdo_num}"
        _, rdo_value = self.send_cmd(cmd)
        rdo_value = int(rdo_value)
        return rdo_value

    def set_rdo(self, rdo_num: int, val: bool):
        """Sets RDO value.

        Args:
            rdo_num (int): RDO number.
            val (bool): Value.
        """
        if val:
            val = "true"
        else:
            val = "false"
        cmd = f"setrdo:{rdo_num}:{val}"
        self.send_cmd(cmd)

    def set_sys_var(self, sys_var: str, val: bool):
        """Sets system variable to True or False.

        Args:
            sys_var (str): System variable name.
            val (bool): Value.
        """
        if val:
            val = "T"
        else:
            val = "F"
        cmd = f"setsysvar:{sys_var}:{val}"
        self.send_cmd(cmd)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fanucpy']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.22.3,<2.0.0', 'scipy>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'fanucpy',
    'version': '0.1.11',
    'description': 'Python package for FANUC industrial robots',
    'long_description': '# fanucpy: Python package for FANUC industrial robots\n\n## Acknowledgements\nThis work was developed at the [Institute for Advanced Manufacturing at the University of Nottingham](https://www.nottingham.ac.uk/ifam/index.aspx) as a part of the [Digital Manufacturing and Design Training Network](https://dimanditn.eu/).\n\nThis project has received funding from the European Union’s Horizon 2020 research and innovation programme under the Marie Skłodowska-Curie grant agreement No 814078.\n\n## Software contents\nThe package consists of two parts: \n1. Robot interface code written in Python programming language\n2. FANUC robot controller driver (tested with R-30iB Mate Plus Controller) written in KAREL and FANUC teach pendant languages\n\nThe communication protocol between the Python package and the FANUC robot controller is depicted below:\n![Communication Protocol](https://github.com/torayeff/fanucpy/raw/main/media/CommProtocol.png)\n\n## Python package installation\n```bash\npip install fanucpy\n```\n\n## Driver installation\nFollow these [steps](https://github.com/torayeff/fanucpy/blob/main/fanuc.md) to install FANUC driver.\n\n## Usage\n### Connect to a robot:\n```python\nfrom fanucpy import Robot\n\nrobot = Robot(\n    robot_model="Fanuc",\n    host="192.168.1.100",\n    port=18735,\n    ee_DO_type="RDO",\n    ee_DO_num=7,\n)\n\nrobot.connect()\n```\n\n### Moving\n```python\n# move in joint space\nrobot.move(\n    "joint",\n    vals=[19.0, 66.0, -33.0, 18.0, -30.0, -33.0],\n    velocity=100,\n    acceleration=100,\n    cnt_val=0,\n    linear=False\n)\n\n# move in cartesian space\nrobot.move(\n    "pose",\n    vals=[0.0, -28.0, -35.0, 0.0, -55.0, 0.0],\n    velocity=50,\n    acceleration=50,\n    cnt_val=0,\n    linear=False\n)\n```\n\n### Opening/closing gripper\n```Python\n# open gripper\nrobot.gripper(True)\n\n# close gripper\nrobot.gripper(False)\n```\n\n### Querying robot state\n```python\n# get robot state\nprint(f"Current pose: {robot.get_curpos()}")\nprint(f"Current joints: {robot.get_curjpos()}")\nprint(f"Instantaneous power: {robot.get_ins_power()}")\nprint(f"Get gripper state: {robot.get_rdo(7)}")\n```\n\n### Calling external program\n```python\nrobot.call_prog(prog_name)\n```\n\n### Get/Set RDO\n```python\nrobot.get_rdo(rdo_num=7)\nrobot.set_rdo(rdo_num=7, value=True)\n```\n\n## Contributions\nExternal contributions are welcome!\n\n- Agajan Torayev: Key developer\n- Fan Mo: Support with documentation\n- Michael Yiu: External contributor\n\n\n## RobotApp\nWe introduce an experimental feature: Robot Apps. This class facilitates modularity and plug-and-produce functionality. Check the following example apps:\n\n1. [Pick and Place App](examples/PickAndPlaceApp.py)\n1. [Aruco Tracking App](examples/ArucoTrackingApp.py)\n1. [FANUC ChatGPT](examples/fanucpy-gpt/README.MD)\n\n## Citation\nPlease use the following to cite if you are using this library in academic publications [Towards Modular and Plug-and-Produce Manufacturing Apps](https://www.sciencedirect.com/science/article/pii/S2212827122004255)\n```\n@article{torayev2022towards,\n  title={Towards Modular and Plug-and-Produce Manufacturing Apps},\n  author={Torayev, Agajan and Mart{\\\'\\i}nez-Arellano, Giovanna and Chaplin, Jack C and Sanderson, David and Ratchev, Svetan},\n  journal={Procedia CIRP},\n  volume={107},\n  pages={1257--1262},\n  year={2022},\n  publisher={Elsevier}\n}\n```\n',
    'author': 'Agajan Torayev',
    'author_email': 'torayeff@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/torayeff/fanucpy',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.11',
}


setup(**setup_kwargs)
